# © Copyright Databand.ai, an IBM Company 2022

import contextlib
import functools
import logging

from typing import List

import attr
import psycopg2
import sqlparse

from dbnd import log_dataset_op
from dbnd._core.log.external_exception_logging import log_exception_to_server
from dbnd._core.utils.sql_tracker_common.sql_extract import (
    READ,
    WRITE,
    SqlQueryExtractor,
)
from dbnd_redshift.sdk.redshift_connection_collection import (
    PostgresConnectionRuntime,
    RedshiftConnectionCollection,
)
from dbnd_redshift.sdk.redshift_connection_extractor import get_redshift_dataset
from dbnd_redshift.sdk.redshift_utils import (
    copy_to_temp_table,
    get_last_query_records_count,
)
from dbnd_redshift.sdk.redshift_values import RedshiftOperation
from dbnd_redshift.sdk.wrappers import (
    DbndConnectionWrapper,
    DbndCursorWrapper,
    PostgresConnectionWrapper,
)


REDSHIFT_TRACKER_OP_SOURCE = "redshift_tracker"

logger = logging.getLogger(__name__)


@attr.s
class RedshiftTrackerConfig:
    """
    Config class for redshift tracker, copycats log_dataset_op args

    Attributes:
        with_histograms(bool): Should calculate histogram of the given data - relevant only with data param.
            - Boolean to calculate or not on all the data columns.
        with_stats(bool): Should extract schema of the data as meta-data of the target - relevant only with data param.
            - Boolean to calculate or not on all the data columns.
        with_partition(bool): If True, the webserver tries to detect partitions of our datasets and extract them from the path,
                        otherwise not manipulating the dataset path at all.
        with_preview(bool): Should extract preview of the data as meta-data of the target - relevant only with data param.
        with_schema(bool): Should extract schema of the data as meta-data of the target - relevant only with data param.
        send_metrics(bool): Should report preview, schemas and histograms as metrics.
    """

    with_preview: bool = attr.ib(default=False)
    with_size: bool = attr.ib(default=None)
    with_schema: bool = attr.ib(default=True)
    with_stats: bool = attr.ib(default=False)
    with_histograms: bool = attr.ib(default=None)
    send_metrics: bool = attr.ib(default=True)
    with_partition: bool = attr.ib(default=None)
    with_percentiles: bool = attr.ib(default=True)

    def __attrs_post_init__(self):
        if self.with_stats and not self.with_schema:
            logger.warning(
                "Column level stats require schema extraction, ignoring with_schema=False"
            )
            self.with_schema = True


class RedshiftTracker:
    def __init__(
        self,
        calculate_file_path=None,
        conf: RedshiftTrackerConfig = RedshiftTrackerConfig(),
    ):

        self.conf = conf

        self.connections: RedshiftConnectionCollection[
            int, PostgresConnectionRuntime
        ] = RedshiftConnectionCollection(lambda: PostgresConnectionRuntime(None, []))

        # custom function for file path calculation
        self.calculate_file_path = calculate_file_path
        self.dataframe = None  # type: Pandas.DF

    def __enter__(self):
        if not hasattr(psycopg2.connect, "__dbnd_patched__"):
            connect_original = psycopg2.connect

            @functools.wraps(connect_original)
            def redshift_connect(*args, **kwargs):
                original_connection = connect_original(*args, **kwargs)
                return DbndConnectionWrapper(original_connection)

            redshift_connect.__dbnd_patched__ = connect_original
            psycopg2.connect = redshift_connect

        if not hasattr(psycopg2.extensions.register_type, "__dbnd_patched__"):
            register_type_original = psycopg2.extensions.register_type

            @functools.wraps(register_type_original)
            def psycopg_register_type(*args, **kwargs):
                try:
                    if isinstance(args[1], DbndConnectionWrapper):
                        connection = args[1]._sqla_unwrap
                    else:
                        connection = args[1]
                    return register_type_original(args[0], connection, **kwargs)
                except Exception:
                    logger.exception(
                        "Error patching psycopg register_type, using original method"
                    )
                    return register_type_original(*args, **kwargs)

            psycopg_register_type.__dbnd_patched__ = register_type_original
            psycopg2.extensions.register_type = psycopg_register_type

        if not hasattr(DbndCursorWrapper.execute, "__dbnd_patched__"):
            execute_original = DbndCursorWrapper.execute

            @functools.wraps(execute_original)
            def redshift_cursor_execute(cursor_self, *args, **kwargs):
                with self.track_execute(cursor_self, *args, **kwargs):
                    return execute_original(cursor_self, *args, **kwargs)

            redshift_cursor_execute.__dbnd_patched__ = execute_original
            DbndCursorWrapper.execute = redshift_cursor_execute

        if not hasattr(DbndConnectionWrapper.close, "__dbnd_patched__"):
            close_original = DbndConnectionWrapper.close

            @functools.wraps(close_original)
            def redshift_connection_close(connection_self, *args, **kwargs):
                conn = self.connections.get_connection(
                    connection_self, PostgresConnectionWrapper(connection_self)
                )
                self.flush_operations(conn)

                return close_original(connection_self, *args, **kwargs)

            redshift_connection_close.__dbnd_patched__ = close_original
            DbndConnectionWrapper.close = redshift_connection_close

        return self

    @staticmethod
    def unpatch_method(obj, original_attr, patched_attr="__dbnd_patched__"):
        method = getattr(obj, original_attr)
        if hasattr(method, patched_attr):
            setattr(obj, original_attr, getattr(method, patched_attr))

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.unpatch_method(DbndCursorWrapper, "execute")
        self.unpatch_method(DbndConnectionWrapper, "close")
        self.unpatch_method(psycopg2, "connect")
        self.unpatch_method(psycopg2.extensions, "register_type")

        for connection in self.connections.values():
            self.flush_operations(connection.connection)

    def flush_operations(self, connection: PostgresConnectionWrapper):
        if connection in self.connections:
            for op in self.connections.get_operations(connection):

                if self.conf.with_schema:
                    op.extract_schema(connection)
                    if self.conf.with_stats:
                        op.extract_stats(
                            connection, with_percentiles=self.conf.with_percentiles
                        )
                if self.conf.with_preview:
                    op.extract_preview(connection)

                log_dataset_op(
                    op_path=op.render_connection_path(connection),
                    op_type=op.op_type,
                    success=op.success,
                    data=op,
                    error=op.error,
                    with_preview=self.conf.with_preview,
                    send_metrics=self.conf.send_metrics,
                    with_schema=self.conf.with_schema,
                    with_partition=self.conf.with_partition,
                    with_stats=self.conf.with_stats,
                    with_histograms=self.conf.with_histograms,
                    operation_source=REDSHIFT_TRACKER_OP_SOURCE,
                )
            # we clean all the batch of operations we reported so we don't report twice
            self.connections.clear_operations(connection)

    def set_dataframe(self, dataframe):
        """
        set dataframe
        Args:
            dataframe(pandas.DF): data structure with columns Which represents the information we have read
        """
        try:
            import pandas as pd

            if isinstance(dataframe, pd.DataFrame):
                self.dataframe = dataframe
            else:
                logger.exception(
                    "Error occurred during set dataframe. provided dataframe is not valid"
                )
        except Exception as e:
            logger.exception("Error occurred during set dataframe: %s", self.dataframe)
            log_exception_to_server(e)

    @contextlib.contextmanager
    def track_execute(self, cursor, command, *args, **kwargs):
        if cursor not in self.connections:
            self.connections.new_connection(
                cursor,
                PostgresConnectionRuntime(
                    PostgresConnectionWrapper(cursor.connection), []
                ),
            )
        success = True
        error = None
        try:
            yield
        except Exception as e:
            success = False
            error = str(e)
            raise
        finally:
            try:
                operations = build_redshift_operations(
                    cursor,
                    command,
                    success,
                    self.calculate_file_path,
                    error,
                    self.dataframe,
                )
                if operations:
                    first_op = operations[0]
                    if first_op.expect_tmp_table(self.conf):
                        copy_to_temp_table(
                            cursor.connection,
                            first_op.schema_name,
                            first_op.table_name,
                            command,
                        )
                    # Only extend self.connections obj operations
                    # if read or write operation occurred in command
                    self.connections.add_operations(cursor, operations)

            except Exception as e:
                logging.exception("Error parsing redshift query")
                log_exception_to_server(e)


def build_redshift_operations(
    cursor: DbndCursorWrapper,
    command: str,
    success: bool,
    calculate_file_path,
    error: str,
    dataframe=None,  # type: pd.DataFrame
) -> List[RedshiftOperation]:
    operations = []
    if calculate_file_path:
        sql_query_extractor = SqlQueryExtractor(calculate_file_path)
    else:
        sql_query_extractor = SqlQueryExtractor()
    clean_command = sql_query_extractor.clean_query(command)
    # find the relevant operations schemas from the command
    parsed_query = sqlparse.parse(clean_command)[0]
    extracted = sql_query_extractor.extract_operations_schemas(parsed_query)

    if not extracted:
        # This is DML statement and no read or write occurred
        return operations

    redshift_dataset = get_redshift_dataset(cursor.connection, clean_command)

    source_name = None
    if READ in extracted:
        source_name = list(extracted[READ].values())[0][0].dataset_name

    for op_type, schema in extracted.items():
        if op_type in [READ, WRITE]:
            operation = RedshiftOperation(
                # redshift_connection=cursor.connection,
                records_count=get_last_query_records_count(cursor.connection),
                query=command,
                # TODO: extract query id
                query_id=None,
                success=success,
                error=error,
                dataframe=dataframe,
                schema_name=redshift_dataset.schema,
                database=redshift_dataset.database,
                table_name=redshift_dataset.table,
                host=redshift_dataset.host,
                source_name=source_name,
                extracted_schema=schema,
                dtypes=None,
                op_type=op_type,
            )

            operations.append(operation)

    return operations



# © Copyright Databand.ai, an IBM Company 2022

from collections import defaultdict
from typing import List, Union

import attr

from psycopg2.extensions import connection

from dbnd_redshift.sdk.redshift_values import RedshiftOperation
from dbnd_redshift.sdk.wrappers import (
    DbndConnectionWrapper,
    DbndCursorWrapper,
    PostgresConnectionWrapper,
)


def get_object_hash(obj):
    connection_hash = None
    if (
        isinstance(obj, DbndCursorWrapper)
        or isinstance(obj, DbndConnectionWrapper)
        or isinstance(obj, PostgresConnectionWrapper)
    ):
        connection_hash = hash(obj.connection)
    elif isinstance(obj, connection):
        connection_hash = hash(connection)

    return connection_hash


@attr.s(auto_attribs=True)
class PostgresConnectionRuntime:
    connection: PostgresConnectionWrapper
    operations: List[RedshiftOperation]


KeyTypeUnion = Union[
    DbndCursorWrapper, DbndConnectionWrapper, PostgresConnectionWrapper, connection
]


class RedshiftConnectionCollection(defaultdict):
    def __init__(self, default_factory=None, **kwargs):
        super().__init__(default_factory, **kwargs)

    def __getitem__(self, item):
        return super().__getitem__(get_object_hash(item))

    def __contains__(self, item):
        return super().__contains__(get_object_hash(item))

    def get(self, key):
        return super().get(get_object_hash(key))

    def new_connection(self, key: KeyTypeUnion, value: PostgresConnectionRuntime):
        """
        Add new connection to map
        :param key: one of KeyTypeUnion union types
        :param value: value for connection, PostgresConnectionRuntime
        :return: None
        """
        hashed_key = get_object_hash(key)
        if hashed_key:
            super().__setitem__(hashed_key, value)

    def get_connection(
        self, key: KeyTypeUnion, default: PostgresConnectionWrapper = None
    ) -> PostgresConnectionWrapper:
        """
        Get psycopg2 connection from selected connection runtime object
        :param key: one of KeyTypeUnion union types
        :param default: default value if doesn't exist in map
        :return: connection object (PostgresConnectionWrapper)
        """
        obj = self.get(key)
        if hasattr(obj, "connection"):
            return obj.connection
        else:
            return default

    def get_operations(self, key: KeyTypeUnion) -> List[RedshiftOperation]:
        """
        Gets operations for given key in map
        :param key: one of KeyTypeUnion union types
        :return: list of sql operations for selected connection
        """
        obj = self.get(key)
        if hasattr(obj, "operations"):
            return obj.operations

    def add_operations(self, key: KeyTypeUnion, operations: List[RedshiftOperation]):
        """
        Given a key and operations adds the operations to the PostgresConnectionRuntime object in the given key
        :param key: one of KeyTypeUnion union types
        :param operations: list of RedshiftOperation to add
        :return: None
        """
        obj = self.get(key)
        if hasattr(obj, "operations"):
            obj.operations.extend(operations)

    def clear_operations(self, key: KeyTypeUnion):
        """
        Given a key clears operations from the PostgresConnectionRuntime in the dict
        :param key: one of KeyTypeUnion union types
        :return: None
        """
        obj = self.get(key)
        if hasattr(obj, "operations"):
            obj.operations.clear()



# © Copyright Databand.ai, an IBM Company 2022

import re

import attr

from psycopg2.extensions import connection


DEFAULT_REDSHIFT_SCHEMA = "public"


@attr.s
class RedshiftDataset:
    host: str = attr.ib()
    database: str = attr.ib()
    schema: str = attr.ib()
    table: str = attr.ib()


def get_redshift_dataset(conn: connection, query: str) -> RedshiftDataset:
    table, schema = get_target_table_and_schema(conn, query)
    host = get_connection_parameter(conn, "host")
    database = get_connection_parameter(conn, "dbname")
    return RedshiftDataset(host=host, database=database, schema=schema, table=table)


def get_target_table_and_schema(con: connection, query):
    table, schema = _extract_table_and_schema_from_query(query)
    # Schema can also be extracted from search_path provided as a connection option or
    # using `SHOW search_path` query. It can contain more than one schema,
    # hence requires more complex logic.
    if not schema:
        schema = DEFAULT_REDSHIFT_SCHEMA
    return table.lower(), schema.lower()


def _extract_table_and_schema_from_query(query):
    table_and_schema_regex = r"""COPY ["']?((?P<schema>\w+)\.)?(?P<table>\w+)["']?"""
    match = re.search(table_and_schema_regex, query, flags=re.IGNORECASE)
    table = match.group("table")
    schema = match.group("schema")
    return table, schema


def get_connection_parameter(con: connection, parameter: str) -> str:
    return con.info.dsn_parameters[parameter]



# © Copyright Databand.ai, an IBM Company 2022

import logging

from itertools import takewhile
from typing import List, Optional
from urllib.parse import urlparse

import attr
import pandas as pd

from dbnd._core.utils.sql_tracker_common.sql_operation import SqlOperation
from dbnd._core.utils.sql_tracker_common.utils import get_redshift_uri
from dbnd_redshift.sdk.redshift_utils import TEMP_TABLE_NAME, redshift_query
from dbnd_redshift.sdk.wrappers import PostgresConnectionWrapper
from targets.connections import build_conn_path
from targets.providers.pandas.pandas_histograms import PandasHistograms
from targets.value_meta import ColumnStatsArgs, ValueMeta, ValueMetaConf
from targets.values import register_value_type
from targets.values.builtins_values import DataValueType


logger = logging.getLogger(__name__)

NUMERIC_TYPES = [
    "SMALLINT",
    "INT2",
    "INTEGER",
    "INT",
    "INT4",
    "BIGINT",
    "INT8",
    "REAL",
    "FLOAT4",
    "DOUBLE PRECISION",
    "FLOAT8",
    "FLOAT",
    "DECIMAL",
    "NUMERIC",
]


def _strip_quotes_and_lower_case(v: str) -> str:
    without_quotes = v.strip('"') if v else v
    lower_case = without_quotes.lower() if without_quotes else without_quotes
    return lower_case


def _type_conversion(type_name: str) -> str:
    normalized_col_type = "".join(takewhile(lambda c: c != "(", type_name))

    converted_type = type_name

    if normalized_col_type == "character varying":
        converted_type = "str"
    if normalized_col_type == "decimal":
        converted_type = "int"

    return converted_type


def _is_numeric_type(col_type: str):
    return any(num_type in col_type.upper() for num_type in NUMERIC_TYPES)


def get_col_stats_queries(col_name: str, col_type: str, with_percentiles: bool = True):
    most_freq_value_cast = "integer::text" if col_type == "boolean" else "text"

    query_list = [
        f"SELECT '{col_name}' AS name",
        f"(SELECT COUNT(*) FROM {TEMP_TABLE_NAME}) AS row_count",
        f'(SELECT COUNT(*) FROM {TEMP_TABLE_NAME} where "{col_name}" is null) AS null_count',
        f'(SELECT count(distinct "{col_name}") FROM {TEMP_TABLE_NAME}) AS distinct_count',
        f'(SELECT top 1 count("{col_name}") FROM {TEMP_TABLE_NAME} group by "{col_name}" order by count("{col_name}") DESC) AS most_freq_value_count',
        f'(SELECT top 1 "{col_name}" FROM {TEMP_TABLE_NAME} group by "{col_name}" order by count("{col_name}") DESC)::{most_freq_value_cast} AS most_freq_value',
    ]

    if _is_numeric_type(col_type):
        query_list.extend(
            [
                f'(SELECT stddev("{col_name}") FROM {TEMP_TABLE_NAME})::text AS stddev',
                f'(SELECT avg("{col_name}"::float8) FROM {TEMP_TABLE_NAME})::text AS average',  # avg calculates in mem, we're limiting the column datatype so it won't cause numeric overflow
                f'(SELECT min("{col_name}") FROM {TEMP_TABLE_NAME})::text AS minimum',
                f'(SELECT max("{col_name}") FROM {TEMP_TABLE_NAME})::text AS maximum',
            ]
        )
        query_list.extend(_percentile_queries(col_name, with_percentiles))

    else:
        query_list.extend(
            [
                "(SELECT Null) AS stddev",
                "(SELECT Null) AS average",
                "(SELECT Null) AS minimum",
                "(SELECT Null) AS maximum",
            ]
        )
        query_list.extend(_null_queries_for_percentile())

    return query_list


def _percentile_queries(col_name: str, with_percentiles: bool) -> List[str]:
    if not with_percentiles:
        return _null_queries_for_percentile()

    return [
        f'(SELECT percentile_cont(0.25) within group (order by "{col_name}" asc) from {TEMP_TABLE_NAME})::text AS percentile_25',
        f'(SELECT percentile_cont(0.50) within group (order by "{col_name}" asc) from {TEMP_TABLE_NAME})::text AS percentile_50',
        f'(SELECT percentile_cont(0.75) within group (order by "{col_name}" asc) from {TEMP_TABLE_NAME})::text AS percentile_75',
    ]


def _null_queries_for_percentile() -> List[str]:
    return [
        "(SELECT Null) AS percentile_25",
        "(SELECT Null) AS percentile_50",
        "(SELECT Null) AS percentile_75",
    ]


def query_list_to_text(cols_queries_lists: List[List[str]]):
    query = ""
    for col_index, col_query_list in enumerate(cols_queries_lists):
        query += "("
        query += ",\n".join(col_query_list)
        query += ") union all\n" if col_index != len(cols_queries_lists) - 1 else ")\n"

    return query


@attr.s
class RedshiftOperation(SqlOperation):
    database = attr.ib(
        converter=_strip_quotes_and_lower_case, default=None
    )  # type: str
    source_name = attr.ib(
        converter=_strip_quotes_and_lower_case, default=None
    )  # type: str
    host = attr.ib(converter=_strip_quotes_and_lower_case, default=None)  # type: str
    schema_name = attr.ib(
        converter=_strip_quotes_and_lower_case, default=None
    )  # type: str
    table_name = attr.ib(
        converter=_strip_quotes_and_lower_case, default=None
    )  # type: str
    cls_cache = attr.ib(default=None)
    schema_cache = attr.ib(default=None)
    preview_cache: pd.DataFrame = attr.ib(default=None)

    @staticmethod
    def expect_tmp_table(conf):
        return conf.with_stats or conf.with_preview

    def extract_preview(self, connection: PostgresConnectionWrapper):
        """
        Fetches 100 rows from the temporary table for preview
        Args:
            connection: RedShift connection object

        """
        if self.dataframe is not None:
            self.preview_cache = self.dataframe
        elif connection:
            db_preview_query = redshift_query(
                connection.connection, f"SELECT * FROM {TEMP_TABLE_NAME} LIMIT 100"
            )
            if db_preview_query:
                first_row = db_preview_query[0]
                self.preview_cache = pd.DataFrame(
                    db_preview_query, columns=list(first_row.keys())
                )

    def extract_schema(self, connection: PostgresConnectionWrapper):
        """
        Queries redshift for the schema of the table, and formats it in dtypes format
        Args:
            connection: RedShift connection object

        """
        if not connection:
            return

        res_schema = {
            "type": self.__class__.__name__,
            "columns": [],
            "dtypes": {},
            "shape": (),
            "size.bytes": 0,  # TODO: calculate operation size in bytes
        }

        if self.dataframe is not None:
            res_schema.update(
                {
                    "columns": list(self.dataframe.columns),
                    "shape": self.dataframe.shape,
                    "dtypes": {
                        col: str(type_) for col, type_ in self.dataframe.dtypes.items()
                    },
                }
            )
        else:
            if self.table_name is not None:
                redshift_query(
                    connection.connection,
                    f"set search_path to '{self.schema_name}'",
                    fetch_all=False,
                )

                desc_results = redshift_query(
                    connection.connection,
                    f"select * from pg_table_def where tablename='{self.table_name}'",
                )

                if desc_results:
                    for col_desc in desc_results:
                        if len(col_desc) > 2:
                            res_schema["columns"].append(col_desc[2])
                            res_schema["dtypes"][col_desc[2]] = _type_conversion(
                                col_desc[3]
                            )

                res_schema["shape"] = (self.records_count, len(res_schema["columns"]))

                if desc_results is not None:
                    self.schema_cache = res_schema

    def extract_stats(
        self, connection: PostgresConnectionWrapper, with_percentiles: bool
    ):
        """
        Extracts column level stats of data in motion, data is copied to a temp table in redshift (in redshift_tracker)
        and here it is queried to extract column level statistics from it
        The query is constructed with respect to the schema of the table, int values calculates (std, min, max, etc...)
        on top of the normally computed column level stats (nullity, row_count, distinct, frequent values, etc...)

        Example:
            value = RedshiftOperation(...)
            value.compute_stats(redshift_connection)

        """
        stats = []  # type: List[ColumnStatsArgs]

        if not connection:
            logger.warning("No redshift connection, can not extract column level stats")
            return

        if self.schema:
            queries = [
                list(
                    get_col_stats_queries(
                        col, col_type, with_percentiles=with_percentiles
                    )
                )
                for col, col_type in self.schema["dtypes"].items()
            ]

            query = query_list_to_text(queries)

            if not query:
                logger.warning(
                    "redshift_tracker: cannot execute empty query for column level stats"
                )
                return

            column_stats = redshift_query(connection.connection, query)

            if column_stats:
                for col in column_stats:
                    null_percentage = 0
                    if col["row_count"] > 0:
                        null_percentage = (col["null_count"] / col["row_count"]) * 100
                    stats.append(
                        ColumnStatsArgs(
                            column_name=col["name"],
                            column_type=self.schema["dtypes"][col["name"]],
                            records_count=col["row_count"],
                            mean_value=col["average"],
                            min_value=col["minimum"],
                            max_value=col["maximum"],
                            std_value=col["stddev"],
                            quartile_1=col["percentile_25"],
                            quartile_2=col["percentile_50"],
                            quartile_3=col["percentile_75"],
                            distinct_count=col["distinct_count"],
                            null_count=col["null_count"],
                            non_null_count=col["row_count"] - col["null_count"],
                            null_percent=null_percentage,
                            unique_count=col["distinct_count"],
                            most_freq_value=col["most_freq_value"],
                            most_freq_value_count=col["most_freq_value_count"],
                        )
                    )

            self.cls_cache = stats

    def render_connection_path(self, connection) -> str:
        if self.is_file:
            urlparsed = urlparse(self.source_name)
            return build_conn_path(
                conn_type=urlparsed.scheme,
                hostname=urlparsed.netloc,
                port=urlparsed.port,
                path=urlparsed.path,
            )
        else:
            return get_redshift_uri(
                self.host, self.database, self.schema_name, self.table_name
            )

    @property
    def columns(self) -> Optional[List[str]]:
        try:
            return list(self.schema["columns"])
        except TypeError:
            return None

    @property
    def columns_count(self) -> Optional[int]:
        try:
            return len(self.schema["dtypes"])
        except TypeError:
            return None

    @property
    def preview(self):
        if self.preview_cache is not None:
            return self.preview_cache.to_string(index=False, max_rows=10)
        return None

    @property
    def schema(self):
        return self.schema_cache

    @property
    def column_stats(self):
        return self.cls_cache


class RedshiftTableValueType(DataValueType):
    type = RedshiftOperation

    support_merge = False
    is_lazy_evaluated = True

    def get_value_meta(self, value: RedshiftOperation, meta_conf: ValueMetaConf):

        # currently, histograms are not supported
        histograms = {}
        hist_sys_metrics = None

        dimensions = (value.records_count, None)

        data_schema = None
        if meta_conf.log_schema:
            data_schema = value.schema
            dimensions = value.schema["shape"]

        column_stats = {}
        if meta_conf.log_stats:
            if value.dataframe is not None:
                column_stats, _ = PandasHistograms(
                    value.dataframe, meta_conf
                ).get_histograms_and_stats()
            else:
                column_stats = value.column_stats

        preview = ""
        if meta_conf.log_preview:
            preview = value.preview

        return ValueMeta(
            value_preview=preview,
            data_dimensions=dimensions,
            data_schema=data_schema,
            data_hash=str(hash(self.to_signature(value))),
            columns_stats=column_stats,
            histogram_system_metrics=hist_sys_metrics,
            histograms=histograms,
            query=value.query,
        )


register_value_type(RedshiftTableValueType())



# © Copyright Databand.ai, an IBM Company 2022

import logging
import re
import sys
import traceback

from psycopg2.extensions import connection as psycopg2_connection
from psycopg2.extras import DictCursor

from dbnd._core.log.external_exception_logging import log_exception_to_server


logger = logging.getLogger(__name__)


COPY_ROWS_COUNT_QUERY = "select pg_last_copy_count();"

TEMP_TABLE_NAME = "DBND_TEMP"


def redshift_query(
    connection: psycopg2_connection, query: str, params=None, fetch_all=True
):
    if connection is not None and not connection.closed:
        with connection.cursor(cursor_factory=DictCursor) as cursor:
            try:
                cursor.execute(query, params)
                if fetch_all:
                    return cursor.fetchall()

            except Exception as e:
                cursor.execute("ROLLBACK")
                traceback.print_exc(file=sys.stdout)
                logger.exception(
                    f"Error occurred during querying redshift, query: {query}"
                )
                log_exception_to_server(e)
    else:
        logger.exception("Error with redshift connection")


def get_last_query_records_count(connection: psycopg2_connection):
    """
    Returns the number of rows that were loaded by the last COPY command run in the current session.
    """
    result_set = redshift_query(connection, COPY_ROWS_COUNT_QUERY)
    if result_set and len(result_set) == 1:
        dimensions = result_set[0]
        if dimensions[0]:
            return dimensions[0]


def build_schema_from_dataframe(dataframe):
    if dataframe is not None:
        try:
            df_schema = dataframe.dtypes.to_dict()
            df_schema = dict((k, str(v)) for k, v in df_schema.items())
        except Exception as e:
            df_schema = None
            logger.exception(
                "Error occurred during build schema from dataframe: %s", dataframe
            )
            log_exception_to_server(e)
    else:
        df_schema = None
    return df_schema


def copy_to_temp_table(redshift_connection, schema_name, table_name, query):
    full_table_name = (
        f"{schema_name}.{table_name}"
        if schema_name is not None and schema_name != "public"
        else table_name
    )
    if table_name is not None:
        redshift_query(
            redshift_connection,
            f"CREATE TEMP TABLE {TEMP_TABLE_NAME} (LIKE {full_table_name});",
            fetch_all=False,
        )
        copy_tmp_table_query = re.sub(
            re.escape(full_table_name),
            TEMP_TABLE_NAME,
            query,
            flags=re.IGNORECASE,
            count=1,
        )
        redshift_query(redshift_connection, copy_tmp_table_query, fetch_all=False)
        redshift_query(redshift_connection, "COMMIT", fetch_all=False)
    else:
        logger.error("Couldn't extract table name")
